import {AfterViewInit, Component, Input, OnInit, ViewChild, ViewContainerRef} from '@angular/core';
import {MapMarker, MapWidget} from "../models/map-widget";
import * as LF from 'leaflet';
import {UserService} from "../../../../services/user.service";
import {Community} from "../../../../models/Community";
import {UtilsService} from "../../../../services/utils.service";
import {TranslateService} from "@ngx-translate/core";
import {LeafletMouseEvent, Marker} from "leaflet";
import {forkJoin} from "rxjs";
import {DataService} from "../../../../services/data.service";
import {DatasourceConfiguration} from "../../../Widget";

@Component({
    selector: 'map-widget',
    templateUrl: './map-widget.component.html',
    styleUrls: ['./map-widget.component.css']
})
export class MapWidgetComponent implements OnInit, AfterViewInit {
    @Input() widget?: MapWidget;
    @ViewChild('popupContainer', { read: ViewContainerRef }) popupContainer?: ViewContainerRef;
    private map: any;
    loadingData: boolean = false;

    communities: Community [] = [];
    markers: LF.Marker[] = [];

    constructor(
        public userService: UserService,
        private utils: UtilsService,
        private dataService: DataService,
        private translateService: TranslateService
    ) {
    }

    ngOnInit(): void {}

    ngAfterViewInit(): void {
        this.initCommunities();
    }

    private initCommunities() {
        this.userService.getCommunities().subscribe({
            next: result => {
                this.communities = result;
                this.initMap();
                this.initMarkers();
            },
            error: err  => this.utils.showErrorMessage(err)
        })
    }

    initMap() {
      this.map = LF.map('map').setView([44.498955, 11.327591], 10);
      LF.tileLayer('https://tile.openstreetmap.org/{z}/{x}/{y}.png', {
          maxZoom: 19,
          attribution: '&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>'
      }).addTo(this.map);
    }

    private initMarkers() {
        if (this.map && this.widget?.markers) {
            this.widget.markers.forEach(markerConfig => {
                const isUserCommunity = markerConfig && markerConfig.communityid === "my_communities";
                if (isUserCommunity) {
                    if (this.userService.user?.usercommunities && this.userService.user?.usercommunities.length > 0) {
                        this.userService.user?.usercommunities.forEach(uc => {
                            this.createMarkerAndAddToMap(uc, markerConfig);
                        })
                    }
                    else {
                        this.utils.showErrorMessage("MAP_WIDGET.USER_HAS_NOT_ANY_COMMUNITY");
                    }
                }
                else {
                    const community = this.communities.find(o => o.communityid == markerConfig.communityid);
                    if (community) {
                       this.createMarkerAndAddToMap(community, markerConfig);
                    }
                    else {
                        this.utils.showErrorMessage("MAP_WIDGET.COMMUNITY_NOT_FOUND")
                    }
                }
            });
            this.fitMarkers();
        }
    }

    private createMarkerAndAddToMap(uc: Community, config: MapMarker) {
        const marker = LF.marker([
            uc.georeferencex || 0,
            uc.georeferencey || 0,

        ]);

        marker.bindTooltip(uc.communityName || "", {
            permanent: false, // Tooltip shows on hover by default
            direction: 'top', // Position the tooltip above the marker
        });

        marker.on('click', () => {
            if (!this.loadingData) {
                const content = this.getHtmlMessage("Caricando i dati ...");

                if (marker.getPopup()) {
                    marker.setPopupContent(content);
                }
                else {
                    marker.bindPopup(content);
                    marker.openPopup();
                }

                this.loadData(marker, config, {"communityid": uc.communityid});
            }
        });
        marker.addTo(this.map);
        this.markers.push(marker);
    }

    private loadData(marker: Marker, config: MapMarker, parameterValues: any) {
        this.loadingData = true;

        let httpPosts = [];

        for (const kpi of config.kpis || []) {
            httpPosts.push(this.dataService.getData(kpi, parameterValues));
        }

        const responses: any[] = [];
        forkJoin(httpPosts).subscribe({
            next: result => {
                for (let res of result) {
                    let response = {
                        "result": res?.result || [],
                        "metadata": res?.metadata || {}
                    };
                    responses.push(response);
                }
                this.loadingData = false;
                this.buildMarkerHtml(marker, responses, config);
            },
            error: er => {
                this.loadingData = false;
                marker.setPopupContent(this.getHtmlMessage("Error during data loading!"));
            }
        });
    }

    private buildMarkerHtml(marker: Marker, data: any[], config: MapMarker) {
        if (data) {
            marker.setPopupContent(this.getHtmlMessage(JSON.stringify(data)));
            let body = "";
            data.forEach((d, index) => {
                const kpi = config.kpis![index];
                body += this.buildKpiHtml(d, kpi);
            });

            if (body) {
                marker.setPopupContent(body);
            }
            else {
                this.translateService.getTranslation("MAP_WIDGET.NO_DATA").subscribe(o => {
                    marker.setPopupContent(this.getHtmlMessage(o));
                });
            }
        }
        else {
            this.translateService.getTranslation("MAP_WIDGET.NO_DATA").subscribe(o => {
                marker.setPopupContent(this.getHtmlMessage(o));
            });
        }
    }

    private buildKpiHtml(d: any, kpi: DatasourceConfiguration):string {
        // d: {result: [], metadata: [{dbcolumnName: "column name"}.....]}
        const header = `<h4>${kpi.parametricQuery?.name}</h4>`
        return header + this.buildBodyHtml(d.metadata, d.result);
    }

    private buildBodyHtml(metadata: any[], datas: any[]): string {
        let result = ""
        metadata.forEach((column, index) => {
            if (column.name) {
                console.info("column: ", column);
                result += `<div style="font-weight: bold;">${column?.description} </div>`;
                datas.forEach(data => {
                    result += this.getDataHtml(data, column);
                });
            }
        })
        return result;
    }

    private getDataHtml(data: any,   column: any) {
        let result = `<div style="font-style: italic;">${data ? data[column.name] + ' ' + (column?.unitOfMeasure ? column?.unitOfMeasure : '')   : '-'}</div>`;

        if (column?.type === 'boolean') {
            result = `<div style="width: 1rem; height: 1rem; border-radius: .6rem; background-color: ${data[column.name] ? 'green' : 'red'}"></div>`;
        }

        return result;
    }

    private getHtmlMessage(message: string) {
        return `<div style="width: 100%; display: flex; flex-direction: row; align-items: center; justify-content: space-between">
            <div>${message}</div>
        </div>`;
    }

    private fitMarkers(): void {
        const group = LF.featureGroup(this.markers);
        this.map.fitBounds(group.getBounds());
    }
}
